#
# Copyright 2024 Ettus Research, a National Instruments Brand
#
# SPDX-License-Identifier: GPL-3.0-or-later
#
""" @package usrp_clock
Python UHD module containing the MultiUSRPClock
"""

from .multi_usrp_clock import MultiUSRPClock
