"""Image builder: Common definitions.

Copyright 2019 Ettus Research, A National Instrument Brand

SPDX-License-Identifier: GPL-3.0-or-later
"""

# Supported protocol version
RFNOC_PROTO_VERSION = "1.0"

DEVICE_NAME = "_device_"

NONE_PORT = "_none_"
